package view;

import javafx.scene.Group;
import javafx.scene.shape.*;
import model.Board;
import model.Rook;
import model.Knight;
import model.Bishop;
import model.Queen;
import model.King;
import model.Piece;
import model.Position;

import java.util.ArrayList;

public class BoardView {

	public static final int BOARD_SIZE = 16, CENTERED_ORIGIN = -BOARD_SIZE*7/2, BOARD_HEIGHT = -5, GRAVEYARD_HEIGHT = -1;

	private final TileView[] tiles; // contains all the squares/tiles of the board (visual)
	private final PieceView[] pieces; // contains all the 3d pieces (visual)
	private int black_death_count, white_death_count; // graveyard counter


	public BoardView(Custom3dModel model, CustomMaterial material, Board board) {
		// creates the board squares, and position them properly
		tiles = new TileView[64];
		for(int j = 1 ; j <= 8 ; j++) {
			for(int i = 1 ; i <= 8 ; i++){
				createTile(new Position(i, j));
			}
		}
		resetColor(material); // init/reset the board squares with black and white

		// creates the 3d pieces (from the board list)
		pieces = new PieceView[64];
		for(int n = 0 ; n < 64 ; n++) {
			pieces[n] = null;
			Piece piece = board.getPiece(n);
			if(piece != null) {
				int type = Custom3dModel.PAWN; // we can't do switch statement on ".getClass()"
				if (piece.getClass() == Rook.class) { type = Custom3dModel.ROOK; }
				if (piece.getClass() == Knight.class) { type = Custom3dModel.KNIGHT; }
				if (piece.getClass() == Bishop.class) { type = Custom3dModel.BISHOP; }
				if (piece.getClass() == Queen.class) { type = Custom3dModel.QUEEN; }
				if (piece.getClass() == King.class) { type = Custom3dModel.KING; }
				createPiece(piece.getPosition(), piece.getColor(), type, material, model);
			}
		}
		// init graveyards as empty
		black_death_count = 0;
		white_death_count = 0;
	}

	public void updatePiece(Position position) { // updates the piece(i, j), visually, to the right position
		int id = position.getID();
		if(getPiece(id) != null) {
			getPieceObj(id).setTranslateX((position.getX()-1) * BOARD_SIZE + CENTERED_ORIGIN);
			getPieceObj(id).setTranslateZ((position.getY()-1) * BOARD_SIZE + CENTERED_ORIGIN);
		}
	}

	public void updateGraveyardPiece(PieceView piece) { // updates position (visual) of a dead piece
		int x = piece.getPosition().getX();
		int y = piece.getPosition().getY();
		if(x == Position.WHITE_GRAVEYARD) {
			piece.getObj().setTranslateX(-2 * BOARD_SIZE + CENTERED_ORIGIN);
			piece.getObj().setTranslateZ(y * BOARD_SIZE*0.65f - BOARD_SIZE*4);
		} else { // x == Position.BLACK_GRAVEYARD
			piece.getObj().setTranslateX(9 * BOARD_SIZE + CENTERED_ORIGIN);
			piece.getObj().setTranslateZ(-y * BOARD_SIZE*0.65f + BOARD_SIZE*4);
		}
		piece.getObj().setTranslateY(GRAVEYARD_HEIGHT); // place at right height (Y axis) on table
	}

	public void createTile(Position position) {
		// creates a tile and with the right size and position
		Box box = new Box(BOARD_SIZE, (double) BOARD_SIZE / 2, BOARD_SIZE);
		box.setTranslateX((position.getX()-1) * BOARD_SIZE + CENTERED_ORIGIN);
		box.setTranslateZ((position.getY()-1) * BOARD_SIZE + CENTERED_ORIGIN);
		// sets the tile on the list
		setTile(new TileView(box, position));
	}

	private void createPiece(Position position, int color, int type, CustomMaterial material, Custom3dModel model) {
		// creates a piece, 3d-model, materials, draw-mode, scale, and the right height (Y-axis)
		MeshView obj = new MeshView(model.get(type));
		obj.setDrawMode(DrawMode.FILL);
		if(color == Piece.WHITE)
			obj.setMaterial(material.get(CustomMaterial.WHITE));
		else
			obj.setMaterial(material.get(CustomMaterial.BLACK));
		int scale = 16; // scales the 3d models to a wanted size, to fit a board tile
		obj.setScaleX(scale);
		obj.setScaleY(-scale); // Y-axis is down (not up) in javaFX, so we need to reverse object
		obj.setScaleZ(-scale); // we reverse Z-axis too, "to keep the orientation of the 3-axis vectors", (reversing Y and Z is equivalent of 180° rotation of X-axis)
		obj.setTranslateY(BOARD_HEIGHT); // place at right 3d height (Y-axis) on board
		// if color is black we rotate
		if(color == Piece.BLACK) {
			obj.setScaleX(-obj.getScaleX());
			obj.setScaleZ(-obj.getScaleZ());
		}
		// sets the piece on the list
		setPiece(new PieceView(obj, position, color));
		// updates piece on the right visual position (corresponding to X-axis and Z-axis placements in 3d)
		updatePiece(position);
	}

	public TileView getTile(int id) {
		return tiles[id];
	}

	public Box getTileObj(int id) {
		return tiles[id].getObj();
	}

	public void setTile(TileView tile) {
		tiles[tile.getPosition().getID()] = tile;
	}

	public PieceView getPiece(int id) {
		return pieces[id];
	}

	public MeshView getPieceObj(int id) {
		return pieces[id].getObj();
	}

	public void setPiece(PieceView piece) {
		pieces[piece.getPosition().getID()] = piece;
	}

	public void resetColor(CustomMaterial material) {
		for(int n = 0, j = 1 ; j <= 8 ; j++) {
			for (int i = 1; i <= 8; i++, n++) {
				// resets board with black and white - bottom left (first case) is black, then alternates
				if((i + j) % 2 == 0) {
					getTileObj(n).setMaterial(material.get(CustomMaterial.BLACK_TILE)); // (visually)
				} else {
					getTileObj(n).setMaterial(material.get(CustomMaterial.WHITE_TILE)); // (visually)
				}
				getTile(n).setState(TileView.NORMAL); // (on 'state' attribute)
			}
		}
	}

	public void setColor(CustomMaterial material, Position position, int type) {
		int id = position.getID();
		// set the new color-type (visually)
		getTileObj(id).setMaterial(material.get(type));
		// save color-type in the "state" attribute of that tile
		switch(type) {
			case CustomMaterial.SELECTED_TILE -> getTile(id).setState(TileView.SELECTED);
			case CustomMaterial.MOVABLE_TILE -> getTile(id).setState(TileView.MOVABLE);
			case CustomMaterial.ATTACK_TILE -> getTile(id).setState(TileView.ATTACK);
			case CustomMaterial.SPECIAL_TILE -> getTile(id).setState(TileView.SPECIAL);
			default -> getTile(id).setState(TileView.NORMAL);
		}
	}

	public void clickPiece(CustomMaterial material, PieceView clickedPiece, Board board) {
		Position position = clickedPiece.getPosition();
		board.setSelected(position);
		resetColor(material);
		setColor(material, position, CustomMaterial.SELECTED_TILE);
		ArrayList<Position> moves = board.getPiece(position).getAvailableMoves(board);
		for (Position move : moves) {
			if(board.isFree(move)) {
				setColor(material, move, CustomMaterial.MOVABLE_TILE);
			} else if(board.isEnemy(move, position)) {
				setColor(material, move, CustomMaterial.ATTACK_TILE);
			} else {
				setColor(material, move, CustomMaterial.SPECIAL_TILE);
			}
			// promotion
			if(board.isPawn(position) && (move.getY() == 8 || move.getY() == 1)) {
				setColor(material, move, CustomMaterial.SPECIAL_TILE);
			}
		}
	}

	public void clickTile(CustomMaterial material, TileView clickedTile, Board board) {
		int state = clickedTile.getState();
		Position selected = board.getSelected();
		Position arrival = clickedTile.getPosition();
		if(state != TileView.NORMAL && state != TileView.SELECTED) {
			if(state == TileView.ATTACK) {
				killPiece(getPiece(arrival.getID()));
			}
			if(state == TileView.SPECIAL) {
				// promotion can be done while attacking too
				if(!board.isFree(arrival)) {
					killPiece(getPiece(arrival.getID()));
				}
				// promotion + rook , to implement
			}
			board.movePiece(selected, arrival); // move piece (on model)
			movePiece(selected, arrival); // move piece (on 3d view)
			resetColor(material);
		}
	}

	public void movePiece(Position origin, Position arrival) {
		PieceView piece = getPiece(origin.getID());
		piece.setPosition(arrival); // changes position to arrival
		setPiece(piece); // copies the piece from origin ID to arrival ID
		pieces[origin.getID()] = null; // removes the origin ID piece
		updatePiece(arrival); // updates 3d model position (visual)
	}

	public void killPiece(PieceView piece) { // graveyards are specified as negative x
		if(piece.getColor() == Piece.WHITE){
			white_death_count++;
			piece.setPosition(new Position(Position.WHITE_GRAVEYARD, white_death_count));
		} else {
			black_death_count++;
			piece.setPosition(new Position(Position.BLACK_GRAVEYARD, black_death_count));
		}
		updateGraveyardPiece(piece);
	}

	public void addAllToScene(Group group) {
		for(int n = 0 ; n < 64 ; n++) {
			if(getTile(n) != null) {
				group.getChildren().addAll(getTileObj(n));
			}
			if(getPiece(n) != null) {
				group.getChildren().addAll(getPieceObj(n));
			}
		}
	}
}
